# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/06_Sparse.ipynb (unless otherwise specified).

__all__ = ['device', 'Mv_scan_py', 'logZ_scan_py', 'ctc_loss_scan_py', 'Mv_scan_cupy', 'logZ_scan', 'cupy_funcs',
           'ctc_loss_scan', 'cupy_func', 'logZ_fwd_cupy', 'fwd_scores_cupy', 'bwd_scores_cupy', 'logZ', 'logZ_bwd_cupy',
           'ctc_loss']

# Cell
from functools import partial, lru_cache as cache
import numpy as np
import cupy as cp
import torch

from .core import semiring, Max, Log
from .utils import *
from .ctc import interleave_blanks, generate_sample_inputs, loss_pytorch, benchmark_fwd_bwd, report, compare_fwd_bwd

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

# Cell
def Mv_scan_py(Ms, idx, v0, S:semiring=Log):
    T, N, C, nz = Ms.shape
    alpha = Ms.new_full((T+1, N, C), S.zero)
    alpha[0] = v0
    for t in range(T):
        alpha[t+1] = S.sum(S.mul(Ms[t], alpha[t, :, idx]), dim=2)
    return alpha

class _LogZ_scan(torch.autograd.Function):
    @staticmethod
    def forward(ctx, Ms, idx, v0, vT, S:semiring, scan):
        alpha = scan(Ms, idx, v0, S)
        ctx.save_for_backward(alpha, Ms, idx, vT)
        ctx.semiring, ctx.scan = S, scan
        return S.sum(S.mul(alpha[-1], vT), dim=1)

    @staticmethod
    def backward(ctx, grad):
        alpha, Ms, idx, vT = ctx.saved_tensors
        S, scan = ctx.semiring, ctx.scan
        T, N, C, nz = Ms.shape
        idx_T = idx.flatten().argsort().reshape(*idx.shape) #transpose
        Ms_T = Ms.reshape(T, N, -1)[:, :, idx_T]
        beta = scan(Ms_T.flip(0), idx_T // nz, vT, S)
        g = S.mul(S.mul(Ms.reshape(T, N, -1), alpha[:-1, :, idx.flatten()]).reshape(T, N, C, nz), beta[:-1, :, :, None].flip(0))
        g = S.dsum(g.reshape(T, N, -1), dim=2).reshape(T, N, C, nz)
        return grad[None, :, None, None] * g, None, None, None, None, None

def logZ_scan_py(Ms, idx, v0, vT, S:semiring):
    return _LogZ_scan.apply(Ms, idx, v0, vT, S, Mv_scan_py)

# Cell
from torch.nn.functional import pad

def _ctc_loss(logits, targets, input_lengths, target_lengths, logZ_impl, S:semiring=Log):
    zero, one = [logits.new_full((1,), x) for x in (S.zero, S.one)]
    scores = logits.log_softmax(2)
    states = interleave_blanks(targets, blank_idx=0)
    state_scores = torch.gather(scores, 2, states.expand(scores.size(0), -1, -1))
    final_states = torch.stack([target_lengths*2-1, target_lengths*2], 1)

    T, N, Lp = state_scores.shape
    assert torch.all(input_lengths == T)

    Ms = torch.stack([
        state_scores,
        pad(state_scores[:, :, 1:], (1, 0), value=S.zero),
        pad(torch.where(states[:, 2:] == states[:, :-2], zero.expand(T, N, Lp-2), state_scores[:, :, 2:]), (2, 0), value=S.zero)
    ], -1)

    i = torch.arange(Lp, device=device)
    rot = lambda x, n: torch.cat([x[-n:], x[:-n]])
    idx = torch.stack([i, rot(i, 1), rot(i, 2)], dim=1)

    v0 = torch.cat([one.expand(N, 1), zero.expand(N, Lp - 1)], dim=1)
    vT = zero.expand(N, Lp).clone().scatter_(1, final_states, S.one)

    logZ = logZ_impl(Ms, idx, v0, vT, S)
    return -(logZ / target_lengths).mean()

ctc_loss_scan_py = partial(_ctc_loss, logZ_impl=logZ_scan_py)

# Cell
cupy_funcs = {
    (torch.float32, Log): load_cupy_func('cuda/sparse_scan.cu', 'sparse_Mv_scan', FLOAT='float',  ADD='logsumexp2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float64, Log): load_cupy_func('cuda/sparse_scan.cu', 'sparse_Mv_scan', FLOAT='double',  ADD='logsumexp2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float32, Max): load_cupy_func('cuda/sparse_scan.cu', 'sparse_Mv_scan', FLOAT='float',  ADD='max2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float64, Max): load_cupy_func('cuda/sparse_scan.cu', 'sparse_Mv_scan', FLOAT='double',  ADD='max2', MUL='add', ZERO='{:E}'.format(Log.zero)),
}

def Mv_scan_cupy(Ms, idx, v0, S:semiring):
    T, N, C, nz = Ms.shape
    assert idx.shape == (C, nz)
    alpha = Ms.new_full((T+1, N, C), S.zero)
    alpha[0] = v0
    with cp.cuda.Device(Ms.device.index):
        cupy_funcs[(Ms.dtype, S)](grid=(N, 1, 1), block=(C, 1, 1), shared_mem=2*8*C,
               args=(alpha.data_ptr(), Ms.data_ptr(), idx.to(dtype=torch.int, device=Ms.device).data_ptr(), T, N, C, nz))
    return alpha

def logZ_scan(Ms, idx, v0, vT, S:semiring):
    return _LogZ_scan.apply(Ms, idx, v0, vT, S, Mv_scan_cupy)

ctc_loss_scan = partial(_ctc_loss, logZ_impl=logZ_scan)

# Cell
@cache(None)
def cupy_func(func_name, dtype, S, NZ, K):
    float_types = {torch.float32: 'float', torch.float64: 'double'}
    ops = {
        Log: {'sum': 'logsumexp', 'mul': 'add'},
        Max: {'sum': 'max_', 'mul': 'add'},
    }
    fname = 'cuda/sparse_logZ.cu'
    return load_cupy_func(fname, func_name, FLOAT=float_types[dtype],  MUL=ops[S]['mul'], ZERO='{:E}'.format(S.zero), SUM=ops[S]['sum'], NZ=NZ, K=K)

def logZ_fwd_cupy(Ms, idx, v0, vT, S:semiring=Log, K=4):
    assert Ms.device.index is not None
    T, N, C, NZ = Ms.shape
    assert idx.shape == (C, NZ)
    idx = idx.to(dtype=torch.int, device=Ms.device)
    Ms_grad = Ms.new_full((T, N, C, NZ), S.zero)
    logZ = Ms.new_full((N, C), S.zero)
    _bytes = 8 if (Ms.dtype == torch.float64) else 4
    with cp.cuda.Device(Ms.device.index):
        cupy_func('logZ_fwd', Ms.dtype, S, NZ, K)(grid=(N, 1, 1), block=(C//K, 1, 1), shared_mem=2*_bytes*C,
               args=(logZ.data_ptr(), Ms_grad.data_ptr(), Ms.data_ptr(), v0.data_ptr(), vT.data_ptr(), idx.data_ptr(), T, N, C))
    return S.sum(logZ, dim=1), Ms_grad

def fwd_scores_cupy(Ms, idx, v0, S:semiring=Log, K=4):
    T, N, C, NZ = Ms.shape
    alphas = Ms.new_full((T+1, N, C), S.zero)
    idx = idx.to(dtype=torch.int, device=Ms.device)
    _bytes = 8 if (Ms.dtype == torch.float64) else 4
    with cp.cuda.Device(Ms.device.index):
        cupy_func('fwd_scores', Ms.dtype, S, NZ, K)(grid=(N, 1, 1), block=(C//K, 1, 1), shared_mem=2*_bytes*C,
               args=(alphas.data_ptr(), Ms.data_ptr(), v0.data_ptr(), idx.data_ptr(), T, N, C))
    return alphas

def bwd_scores_cupy(Ms, idx, vT, S:semiring=Log, K=4):
    T, N, C, NZ = Ms.shape
    betas = Ms.new_full((T+1, N, C), S.zero)
    idx_T = idx.flatten().argsort().to(dtype=torch.int, device=Ms.device) #transpose
    _bytes = 8 if (Ms.dtype == torch.float64) else 4
    with cp.cuda.Device(Ms.device.index):
        cupy_func('bwd_scores', Ms.dtype, S, NZ, K)(grid=(N, 1, 1), block=(C//K, 1, 1), shared_mem=2*_bytes*C,
               args=(betas.data_ptr(), Ms.data_ptr(), vT.data_ptr(), idx_T.data_ptr(), T, N, C))
    return betas

logZ_bwd_cupy = bwd_scores_cupy #backward compatibility for renamed function

class _LogZ(torch.autograd.Function):
    @staticmethod
    def forward(ctx, Ms, idx, v0, vT, S:semiring, K):
        idx = idx.to(device=Ms.device)
        logZ, Ms_grad = logZ_fwd_cupy(Ms, idx, v0, vT, S, K)
        ctx.save_for_backward(Ms_grad, Ms, idx, vT)
        ctx.semiring = S
        ctx.K = K
        return logZ

    @staticmethod
    def backward(ctx, grad):
        Ms_grad, Ms, idx, vT = ctx.saved_tensors
        S, K = ctx.semiring, ctx.K
        T, N, C, NZ = Ms.shape
        betas = bwd_scores_cupy(Ms, idx, vT, S, K=K)
        Ms_grad = S.mul(Ms_grad, betas[1:,:,:,None])
        Ms_grad = S.dsum(Ms_grad.reshape(T, N, -1), dim=2).reshape(T, N, C, NZ)
        return grad[None, :, None, None] * Ms_grad, None, None, None, None, None

def logZ(Ms, idx, v0, vT, S:semiring=Log, K=1):
    return _LogZ.apply(Ms, idx, v0, vT, S, K)

ctc_loss = partial(_ctc_loss, logZ_impl=logZ)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/01_utils.ipynb (unless otherwise specified).

__all__ = ['to_np', 'zero_grad', 'float64', 'compare_fwd_bwd', 'timed', 'benchmark_fwd_bwd', 'report', 'add_checks',
           'load_cupy_func', 'load_cupy_module']

# Cell
import torch
import numpy as np
import cupy as cp
from pathlib import Path

# Cell
def to_np(x):
    return x.clone().detach().cpu().numpy()

def zero_grad(*xs):
    return [x.grad.zero_() for x in xs if hasattr(x, 'grad') and x.grad is not None]

def float64(func):
    return lambda *args: func(*[x.to(torch.float64) if hasattr(x, 'dtype') and x.dtype is torch.float else x for x in args])

def compare_fwd_bwd(impl_A, impl_B, inputs, *args):
    fwds, bwds = [], []
    for impl in (impl_A, impl_B):
        fwd = impl(inputs, *args)
        fwd.backward()
        fwds.append(to_np(fwd))
        bwds.append(to_np(inputs.grad))
        zero_grad(inputs)
    print('fwd diff: {:.2e}'.format(np.max(np.abs(fwds[0]-fwds[1]))))
    print('bwd diff: {:.2e}'.format(np.max(np.abs(bwds[0]-bwds[1]))))
    return fwds, bwds

# Cell
def timed(func, *inputs):
    start, end = [torch.cuda.Event(enable_timing=True) for _ in range(2)]
    start.record(); output = func(*inputs); end.record()
    torch.cuda.synchronize()
    return output, start.elapsed_time(end)

def benchmark_fwd_bwd(fwd_impl, *inputs, warmup=5, nloops=20):
    def fwd_bwd_times(fwd_impl, *inputs):
        output, fwd_time = timed(fwd_impl, *inputs)
        _, bwd_time = timed(output.backward)
        zero_grad(*inputs)
        return (fwd_time, bwd_time)
    [fwd_bwd_times(fwd_impl, *inputs) for _ in range(warmup)]
    fwd_times, bwd_times = map(np.array, zip(*[fwd_bwd_times(fwd_impl, *inputs) for _ in range(nloops)]))
    return {'fwd': fwd_times, 'bwd': bwd_times, 'tot': fwd_times+bwd_times}

def report(times):
    for k,v in times.items():
        print('{!s}: {:.2f}ms ({:.2f}-{:.2f}ms)'.format(k, v.mean(), v.min(), v.max()))

# Cell
def add_checks(raw_kernel):
    def wrapped(grid, block, args, *, shared_mem=0):
        MAX_THREADS = 1024
        if np.prod(block) > MAX_THREADS:
            raise Exception('Block of size {} not allowed. Maximum number of threads allowed per block is {}.'.format(block, MAX_THREADS))
        return raw_kernel(grid, block, args, shared_mem=shared_mem)
    return wrapped

def load_cupy_func(fname, name, **kwargs):
    try: fname = str((Path(__file__).parent / fname).resolve())
    except: pass
    with open(fname) as f:
        code = f.read()
    macros = ['#define {!s} {!s}'.format(k, v) for k,v in kwargs.items()]
    code = '\n'.join(macros + [code])
    return add_checks(cp.RawKernel(code, name))

def load_cupy_module(fname, **kwargs):
    try: fname = str((Path(__file__).parent / fname).resolve())
    except: pass
    with open(fname) as f:
        code = f.read()
    macros = ['#define {!s} {!s}'.format(k, v) for k,v in kwargs.items()]
    code = '\n'.join(macros + [code])
    return cp.RawModule(code=code)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/03_CTC_loss_banded.ipynb (unless otherwise specified).

__all__ = ['device', 'window_start_positions', 'loss_masked', 'prepare_inputs', 'loss_basic', 'fwd_bwd', 'loss_py',
           'loss_cupy', 'cupy_funcs', 'soft_alignments', 'viterbi_alignments']

# Cell
from functools import partial
import torch
import numpy as np
import cupy as cp

from .core import semiring, Log, Max
from .utils import *
from seqdist import ctc

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

# Cell
def window_start_positions(base_alignments, input_lengths, width, num_states):
    t = torch.arange(base_alignments.size(0), device=base_alignments.device)
    base_alignments = torch.where(t[:, None] < input_lengths, base_alignments, num_states - 1)
    return torch.min(torch.max(base_alignments-width//2, torch.zeros_like(base_alignments)), (num_states-width))

def loss_masked(logits, targets, input_lengths, target_lengths, base_alignments, width):
    state_scores, repeat_mask, final_states, input_lengths = ctc.prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths)
    T, N, Lp = state_scores.shape
    window_starts = window_start_positions(base_alignments, input_lengths, width, target_lengths*2 + 1).unsqueeze(-1)
    states = torch.arange(Lp, device=logits.device)
    state_scores = torch.where((window_starts <= states) & (states < window_starts + width), state_scores, logits.new_full((1,), Log.zero))
    logz = ctc._logz_fwd(state_scores, repeat_mask, final_states, input_lengths)
    return -(logz / target_lengths).mean()

# Cell
def prepare_inputs(scores, targets, input_lengths, target_lengths, base_alignments, width:int):
    (T, N, C), device = scores.shape, scores.device
    num_states = target_lengths*2 + 1
    states = ctc.interleave_blanks(targets, blank_idx=0)
    repeat_mask = torch.nn.functional.pad(states[:, 2:] == states[:, :-2], (2, 2), value=0.)
    window_starts = window_start_positions(base_alignments, input_lengths, width, num_states)
    windows = torch.arange(width, device=device) + window_starts.unsqueeze(-1)

    state_scores = torch.gather(scores, 2, states.expand(T, -1, -1).gather(2, windows))
    final_states = (torch.stack([num_states - 2, num_states - 1], 1) - window_starts[input_lengths - 1, torch.arange(N), None])

    return state_scores, repeat_mask, final_states, input_lengths, window_starts

def _logz_fwd(state_scores, repeat_mask, final_states, input_lengths, window_starts, S:semiring=Log):
    T, N, W = state_scores.shape
    zeros = state_scores.new_full((N, W + 2), S.zero)
    alpha = state_scores.new_full((T + 1, N, W), S.zero)
    alpha[0, :, 0] = S.one
    prev_pos = window_starts.new_zeros(N, 1)
    range_W = torch.arange(W + 2, device=state_scores.device)
    for t in range(0, T):
        a = torch.cat([zeros[:, :2], alpha[t], zeros[:, :2]], dim=1)
        a[:, 2:] = S.sum(torch.stack([
                            a[:, 2:], a[:, 1:-1], torch.where(repeat_mask.gather(1, prev_pos + range_W), zeros, a[:, :-2])]), dim=0)
        pos = window_starts[t, :, None]
        alpha[t+1] = S.mul(state_scores[t], a.gather(1, (pos-prev_pos) + range_W[2:]))
        prev_pos = pos
    return S.sum(alpha[input_lengths, torch.arange(N)].gather(1, final_states), dim=1)

def loss_basic(logits, targets, input_lengths, target_lengths, base_alignments, width):
    log_probs = logits.log_softmax(2)
    logz = _logz_fwd(*prepare_inputs(log_probs, targets, input_lengths, target_lengths, base_alignments, width))
    return -(logz / target_lengths).mean()

# Cell
def fwd_bwd(state_scores, repeat_mask, final_states, input_lengths, window_starts, fwd_bwd_impl, S:semiring=Log):
    T, N, W = state_scores.shape
    alpha, beta = [state_scores.new_full((T+1, N, W), S.zero) for _ in range(2)]
    alpha[0, :, 0] = S.one
    beta[input_lengths, torch.arange(N)] = state_scores.new_full((N, W), S.zero).scatter_(1, final_states, S.one)
    window_starts = torch.cat([window_starts.new_zeros(1, N), window_starts], dim=0)
    alpha_T = fwd_bwd_impl(alpha, beta, state_scores, repeat_mask, input_lengths, window_starts, S)
    logz = S.sum(alpha_T.gather(1, final_states), dim=1)
    return alpha, beta, logz

def _fwd_bwd_py(alpha, beta, state_scores, repeat_mask, input_lengths, window_starts, S:semiring=Log):
    T, N, W = state_scores.shape
    zeros = state_scores.new_full((N, W + 2), S.zero)
    range_W = torch.arange(W + 2, device=state_scores.device)
    #fwd
    prev_pos = window_starts[0, :, None]
    for t in range(0, T):
        a = torch.cat([zeros[:, :2], alpha[t], zeros[:, :2]], dim=1)
        a[:, 2:] = S.sum(torch.stack([
                            a[:, 2:], a[:, 1:-1], torch.where(repeat_mask.gather(1, prev_pos + range_W), zeros, a[:, :-2])]), dim=0)
        pos = window_starts[t + 1, :, None]
        alpha[t+1] = S.mul(state_scores[t], a.gather(1, (pos-prev_pos) + range_W[2:]))
        prev_pos = pos
    #bwd
    for t in range(T, 0, -1):
        b = torch.cat([zeros[:, :2], S.mul(beta[t], state_scores[t-1]), zeros[:, :2]], dim=1)
        pos = window_starts[t - 1, :, None]
        b[:, :-2] = S.sum(torch.stack([b[:, :-2], b[:, 1:-1], torch.where(repeat_mask.gather(1, prev_pos + range_W), zeros, b[:, 2:])]), dim=0)
        beta[t-1, t <= input_lengths] = b.gather(1, (pos-prev_pos) + range_W[2:])[t <= input_lengths]
        prev_pos = pos

    return alpha[input_lengths, torch.arange(N)]

class _Logz(torch.autograd.Function):
    @staticmethod
    def forward(ctx, state_scores, repeat_mask, final_states, input_lengths, window_starts, fwd_bwd_impl):
        alpha, beta, logz = fwd_bwd(state_scores, repeat_mask, final_states, input_lengths, window_starts, fwd_bwd_impl, Log)
        ctx.save_for_backward(alpha, beta, input_lengths)
        return logz

    @staticmethod
    def backward(ctx, grad):
        alpha, beta, input_lengths = ctx.saved_tensors
        g = torch.softmax(alpha[1:] + beta[1:], dim=2) * ctc.masked_grad(grad.expand(alpha.size(0)-1, -1), input_lengths)
        return g, None, None, None, None, None

def loss_py(logits, targets, input_lengths, target_lengths, base_alignments, width):
    logz = _Logz.apply(*prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths, base_alignments, width), _fwd_bwd_py)
    return - (logz / target_lengths).mean()

# Cell
cupy_funcs = {
    (torch.float32, Log): load_cupy_func('cuda/ctc_banded.cu', 'fwd_bwd_banded', FLOAT='float',  SUM='logsumexp3', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float64, Log): load_cupy_func('cuda/ctc_banded.cu', 'fwd_bwd_banded', FLOAT='double', SUM='logsumexp3', MUL='add', ZERO='{:E}'.format(Log.zero)),
}

def _fwd_bwd_cupy(alpha, beta, state_scores, repeat_mask, input_lengths, window_starts, S:ctc.semiring):
    T, N, W = state_scores.shape
    _, L = repeat_mask.shape #Note this L is Lp + 2
    alpha_T = torch.empty_like(alpha[0])
    with cp.cuda.Device(state_scores.device.index):
        cupy_funcs[(state_scores.dtype, S)](grid=(N, 2, 1), block=(W, 1, 1), shared_mem=2*8*W,
               args=(alpha_T.data_ptr(), alpha.data_ptr(), beta.data_ptr(), state_scores.data_ptr(), repeat_mask.data_ptr(),
                     input_lengths.data_ptr(), window_starts.data_ptr(), N, L, W))
    return alpha_T

def loss_cupy(logits, targets, input_lengths, target_lengths, base_alignments, width):
    logz = _Logz.apply(*prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths, base_alignments, width), _fwd_bwd_cupy)
    return - (logz / target_lengths).mean()

# Cell
cupy_funcs[(torch.float32, Max)] = load_cupy_func('cuda/ctc_banded.cu', 'fwd_bwd_banded', FLOAT='float',  SUM='max3', MUL='add', ZERO='{:E}'.format(Max.zero))
cupy_funcs[(torch.float64, Max)] = load_cupy_func('cuda/ctc_banded.cu', 'fwd_bwd_banded', FLOAT='double', SUM='max3', MUL='add', ZERO='{:E}'.format(Max.zero))

class _LogzViterbi(torch.autograd.Function):
    @staticmethod
    def forward(ctx, state_scores, repeat_mask, final_states, input_lengths, window_starts, fwd_bwd_impl):
        alpha, beta, logz = fwd_bwd(state_scores, repeat_mask, final_states, input_lengths, window_starts, fwd_bwd_impl, Max)
        ctx.save_for_backward(alpha, beta, input_lengths)
        return logz

    @staticmethod
    def backward(ctx, grad):
        alpha, beta, input_lengths = ctx.saved_tensors
        g = Max.dsum(alpha[1:] + beta[1:], dim=2) * ctc.masked_grad(grad.expand(alpha.size(0) - 1, -1), input_lengths)
        return g, None, None, None, None, None

def soft_alignments(logits, targets, input_lengths, target_lengths, base_alignments, width):
    state_scores, *args = prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths, base_alignments, width)
    _Logz.apply(state_scores.detach_().requires_grad_(), *args, _fwd_bwd_cupy).sum().backward()
    return state_scores.grad

def viterbi_alignments(logits, targets, input_lengths, target_lengths, base_alignments, width):
    state_scores, *args = prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths, base_alignments, width)
    _LogzViterbi.apply(state_scores.detach_().requires_grad_(), *args, _fwd_bwd_cupy).sum().backward()
    return state_scores.grad


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/02_CTC_loss.ipynb (unless otherwise specified).

__all__ = ['device', 'generate_sample_inputs', 'loss_pytorch', 'interleave_blanks', 'prepare_inputs', 'loss_basic',
           'fwd_bwd', 'masked_grad', 'loss_py', 'loss_cupy', 'cupy_funcs', 'viterbi_alignments', 'soft_alignments',
           'loss_direct_cupy', 'Prob']

# Cell
import numpy as np
import cupy as cp
import torch
import torch.nn as nn

from .core import semiring, Log, Max
from .utils import *

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

# Cell
def generate_sample_inputs(T_min, T_max, N, C, L_min, L_max, device=device):
    """
    Args:
        T_min, T_max: bounds on number of time steps
        N: batch size
        C: alphabet size (including blank)
        L_min, L_max: bounds on target length
    """
    logits = torch.randn(T_max, N, C, device=device, requires_grad=True)
    targets = torch.randint(1, C, (N, L_max), dtype=torch.long, device=device)
    input_lengths = torch.randint(T_min, T_max+1, (N,), dtype=torch.long, device=device)
    target_lengths = torch.randint(L_min, L_max+1, (N,), dtype=torch.long, device=device)
    return logits, targets, input_lengths, target_lengths

# Cell
def loss_pytorch(logits, targets, input_lengths, target_lengths, blank=0, reduction='mean', zero_infinity=False):
    log_probs = logits.log_softmax(2)
    return nn.functional.ctc_loss(log_probs, targets, input_lengths, target_lengths, blank, reduction, zero_infinity)

# Cell
def interleave_blanks(targets, blank_idx: int):
    N, L = targets.shape
    interleaved = targets.new_full((N, 2*L+1), blank_idx)
    interleaved[:, 1::2] = targets
    return interleaved

def prepare_inputs(scores, targets, input_lengths, target_lengths):
    states = interleave_blanks(targets, blank_idx=0)
    state_scores = torch.gather(scores, 2, states.expand(scores.size(0), -1, -1))
    repeat_mask = torch.nn.functional.pad(states[:, 2:] == states[:, :-2], (2, 0), value=0.)
    final_states = torch.stack([target_lengths*2-1, target_lengths*2], 1)
    return state_scores, repeat_mask, final_states, input_lengths

def _logz_fwd(state_scores, repeat_mask, final_states, input_lengths, S:semiring=Log):
    T, N, Lp = state_scores.shape
    zeros = state_scores.new_full((N, Lp), S.zero)
    a = state_scores.new_full((N, Lp+2), S.zero)
    a[:, 2] = S.one
    alpha = torch.empty_like(state_scores)
    for t in range(0, T):
        alpha[t] = a[:, 2:] = S.mul(state_scores[t], S.sum(
            torch.stack([a[:, 2:], a[:, 1:-1], torch.where(repeat_mask, zeros, a[:, :-2])]), dim=0))
    return S.sum(alpha[input_lengths-1, torch.arange(N)].gather(1, final_states), dim=1)

def loss_basic(logits, targets, input_lengths, target_lengths):
    log_probs = logits.log_softmax(2)
    logz = _logz_fwd(*prepare_inputs(log_probs, targets, input_lengths, target_lengths))
    return -(logz / target_lengths).mean()

# Cell
def fwd_bwd(state_scores, repeat_mask, final_states, input_lengths, fwd_bwd_impl, S:semiring=Log):
    T, N, Lp = state_scores.shape
    alpha, beta = [state_scores.new_full((T+1, N, Lp), S.zero) for _ in range(2)]
    alpha[0, :, 0] = S.one
    beta[input_lengths, torch.arange(N)] = state_scores.new_full((N, Lp), S.zero).scatter_(1, final_states, S.one)
    alpha_T = fwd_bwd_impl(alpha, beta, state_scores, repeat_mask, input_lengths, S)
    logz = S.sum(alpha_T.gather(1, final_states), dim=1)
    return alpha, beta, logz

def _fwd_bwd_py(alpha, beta, state_scores, repeat_mask, input_lengths, S:semiring=Log):
    T, N, Lp = state_scores.shape
    zeros = alpha.new_full((1,), S.zero)
    #fwd
    a = torch.cat([alpha.new_full((N, 2), S.zero), alpha[0]], 1)
    for t in range(0, T):
        a[:, 2:] = S.mul(state_scores[t], S.sum(torch.stack([a[:, 2:], a[:, 1:-1], torch.where(repeat_mask, zeros, a[:, :-2])]), dim=0))
        alpha[t+1] = a[:, 2:]
    #bwd
    b = alpha.new_full((N, Lp+2), S.zero)
    repeat_mask = torch.cat([repeat_mask[:, 2:], repeat_mask[:, :2]], 1)
    for t in range(T, 0, -1):
        b[:, :-2] = S.mul(beta[t], state_scores[t-1])
        b[:, :-2] = S.sum(torch.stack([b[:, :-2], b[:, 1:-1], torch.where(repeat_mask, zeros, b[:, 2:])]), dim=0)
        beta[t-1, t <= input_lengths] = b[t <= input_lengths, :-2]

    return alpha[input_lengths, torch.arange(N)]

def masked_grad(grad, input_lengths):
    input_mask = (torch.arange(grad.size(0), device=grad.device)[:, None] < input_lengths)
    return torch.where(input_mask, grad, grad.new_zeros((1,))).unsqueeze(2)

class _Logz(torch.autograd.Function):
    @staticmethod
    def forward(ctx, state_scores, repeat_mask, final_states, input_lengths, fwd_bwd_impl, S:semiring):
        alpha, beta, logz = fwd_bwd(state_scores, repeat_mask, final_states, input_lengths, fwd_bwd_impl, S)
        ctx.save_for_backward(alpha, beta, input_lengths)
        ctx.semiring=S
        return logz

    @staticmethod
    def backward(ctx, grad):
        alpha, beta, input_lengths = ctx.saved_tensors
        g = ctx.semiring.dsum(alpha[1:] + beta[1:], dim=2) * masked_grad(grad.expand(alpha.size(0)-1, -1), input_lengths)
        return g, None, None, None, None, None

def loss_py(logits, targets, input_lengths, target_lengths):
    logz = _Logz.apply(*prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths), _fwd_bwd_py, Log)
    return - (logz / target_lengths).mean()

# Cell
cupy_funcs = {
    (torch.float32, Log): load_cupy_func('cuda/ctc.cu', 'fwd_bwd_logspace', FLOAT='float',  SUM='logsumexp3', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float64, Log): load_cupy_func('cuda/ctc.cu', 'fwd_bwd_logspace', FLOAT='double', SUM='logsumexp3', MUL='add', ZERO='{:E}'.format(Log.zero)),
}

def _fwd_bwd_cupy(alpha, beta, state_scores, repeat_mask, input_lengths, S:semiring):
    T, N, Lp = state_scores.shape
    alpha_T = torch.empty_like(alpha[0])
    with cp.cuda.Device(state_scores.device.index):
        cupy_funcs[(state_scores.dtype, S)](grid=(N, 2, 1), block=(Lp, 1, 1), shared_mem=2*8*Lp,
               args=(alpha_T.data_ptr(), alpha.data_ptr(), beta.data_ptr(), state_scores.data_ptr(), repeat_mask.data_ptr(),
                     input_lengths.data_ptr(), N, Lp))
    return alpha_T

def loss_cupy(logits, targets, input_lengths, target_lengths):
    logz = _Logz.apply(*prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths), _fwd_bwd_cupy, Log)
    return - (logz / target_lengths).mean()

# Cell
cupy_funcs[(torch.float32, Max)] = load_cupy_func('cuda/ctc.cu', 'fwd_bwd_logspace', FLOAT='float',  SUM='max3', MUL='add', ZERO='{:E}'.format(Max.zero))
cupy_funcs[(torch.float64, Max)] = load_cupy_func('cuda/ctc.cu', 'fwd_bwd_logspace', FLOAT='double', SUM='max3', MUL='add', ZERO='{:E}'.format(Max.zero))

def viterbi_alignments(logits, targets, input_lengths, target_lengths):
    state_scores, repeat_mask, final_states, input_lengths = prepare_inputs(logits.log_softmax(2), targets, input_lengths, target_lengths)
    _Logz.apply(state_scores.detach_().requires_grad_(), repeat_mask, final_states, input_lengths, _fwd_bwd_cupy, Max).sum().backward()
    return state_scores.grad

def soft_alignments(logits, targets, input_lengths, target_lengths, beta=1.0):
    state_scores, repeat_mask, final_states, input_lengths = prepare_inputs((logits*beta).log_softmax(2), targets, input_lengths, target_lengths)
    _Logz.apply(state_scores.detach_().requires_grad_(), repeat_mask, final_states, input_lengths, _fwd_bwd_cupy, Log).sum().backward()
    return state_scores.grad

# Cell
Prob = semiring(zero=0., one=1., mul=torch.mul, sum=torch.sum, dsum=(lambda x, dim=0: torch.ones_like))
cupy_funcs[(torch.float64, Prob)] = load_cupy_func('cuda/ctc.cu', 'fwd_bwd_logspace', FLOAT='double', SUM='sum3', MUL='mul', ZERO='0.0')

class _LogzDirect(torch.autograd.Function):
    @staticmethod
    def forward(ctx, state_scores, repeat_mask, final_states, input_lengths, fwd_bwd_impl):
        alpha, beta, z = fwd_bwd(state_scores, repeat_mask, final_states, input_lengths, fwd_bwd_impl, Prob)
        ctx.save_for_backward(alpha, beta, state_scores, input_lengths)
        return torch.log(z)

    @staticmethod
    def backward(ctx, grad):
        alpha, beta, state_probs, input_lengths = ctx.saved_tensors
        g = alpha[1:]*beta[1:]
        g = (g/state_probs)*masked_grad(grad.expand(alpha.size(0)-1, -1), input_lengths)/(g.sum(-1, keepdim=True)+1e-38)
        return g, None, None, None, None

def loss_direct_cupy(logits, targets, input_lengths, target_lengths):
    logz = _LogzDirect.apply(*prepare_inputs(logits.softmax(2), targets, input_lengths, target_lengths), _fwd_bwd_cupy)
    return - (logz / target_lengths).mean()


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"max_grad": "00_core.ipynb",
         "semiring": "00_core.ipynb",
         "Log": "00_core.ipynb",
         "Max": "00_core.ipynb",
         "grad": "00_core.ipynb",
         "SequenceDist": "00_core.ipynb",
         "to_np": "01_utils.ipynb",
         "zero_grad": "01_utils.ipynb",
         "float64": "01_utils.ipynb",
         "compare_fwd_bwd": "01_utils.ipynb",
         "timed": "01_utils.ipynb",
         "benchmark_fwd_bwd": "01_utils.ipynb",
         "report": "01_utils.ipynb",
         "add_checks": "01_utils.ipynb",
         "load_cupy_func": "01_utils.ipynb",
         "load_cupy_module": "01_utils.ipynb",
         "device": "06_Sparse.ipynb",
         "generate_sample_inputs": "04_CTC_loss_simple.ipynb",
         "loss_pytorch": "02_CTC_loss.ipynb",
         "interleave_blanks": "02_CTC_loss.ipynb",
         "prepare_inputs": "03_CTC_loss_banded.ipynb",
         "loss_basic": "03_CTC_loss_banded.ipynb",
         "fwd_bwd": "03_CTC_loss_banded.ipynb",
         "masked_grad": "02_CTC_loss.ipynb",
         "loss_py": "03_CTC_loss_banded.ipynb",
         "loss_cupy": "03_CTC_loss_banded.ipynb",
         "cupy_funcs": "06_Sparse.ipynb",
         "viterbi_alignments": "04_CTC_loss_simple.ipynb",
         "soft_alignments": "04_CTC_loss_simple.ipynb",
         "loss_direct_cupy": "02_CTC_loss.ipynb",
         "Prob": "02_CTC_loss.ipynb",
         "window_start_positions": "03_CTC_loss_banded.ipynb",
         "loss_masked": "03_CTC_loss_banded.ipynb",
         "logZ_fwd": "04_CTC_loss_simple.ipynb",
         "dot": "04_CTC_loss_simple.ipynb",
         "LogZ": "05_Normalisation.ipynb",
         "logZ_py": "05_Normalisation.ipynb",
         "mean": "04_CTC_loss_simple.ipynb",
         "logZ_cupy": "04_CTC_loss_simple.ipynb",
         "cupy_funcs_loop": "04_CTC_loss_simple.ipynb",
         "generate_test_example": "05_Normalisation.ipynb",
         "fused_batch_Mv": "05_Normalisation.ipynb",
         "LogZViterbi": "05_Normalisation.ipynb",
         "logz": "05_Normalisation.ipynb",
         "Mv_scan_py": "06_Sparse.ipynb",
         "logZ_scan_py": "06_Sparse.ipynb",
         "ctc_loss_scan_py": "06_Sparse.ipynb",
         "Mv_scan_cupy": "06_Sparse.ipynb",
         "logZ_scan": "06_Sparse.ipynb",
         "ctc_loss_scan": "06_Sparse.ipynb",
         "cupy_func": "06_Sparse.ipynb",
         "logZ_fwd_cupy": "06_Sparse.ipynb",
         "fwd_scores_cupy": "06_Sparse.ipynb",
         "bwd_scores_cupy": "06_Sparse.ipynb",
         "logZ": "06_Sparse.ipynb",
         "logZ_bwd_cupy": "06_Sparse.ipynb",
         "ctc_loss": "06_Sparse.ipynb"}

modules = ["core.py",
           "utils.py",
           "ctc.py",
           "ctc_banded.py",
           "ctc_simple.py",
           "normalisation.py",
           "sparse.py"]

doc_url = "https://davidcpage.github.io/seqdist/"

git_url = "https://github.com/davidcpage/seqdist/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/00_core.ipynb (unless otherwise specified).

__all__ = ['max_grad', 'semiring', 'Log', 'Max', 'grad', 'SequenceDist']

# Cell
import torch
from collections import namedtuple

# Cell
semiring = namedtuple('semiring', ('zero', 'one', 'mul', 'sum', 'dsum'))

def max_grad(x, dim=0):
    return torch.zeros_like(x).scatter_(dim, x.argmax(dim, True), 1.0)

Log = semiring(zero=-1e38, one=0., mul=torch.add, sum=torch.logsumexp, dsum=torch.softmax)
Max = semiring(zero=-1e38, one=0., mul=torch.add, sum=(lambda x, dim=0: torch.max(x, dim=dim)[0]), dsum=max_grad)

# Cell
def grad(f, x):
    x = x.detach().requires_grad_()
    with torch.enable_grad():
        y = f(x)
    return torch.autograd.grad(y, x)[0].detach()

class SequenceDist():
    def __init__(self):
        pass

    def logZ(self, scores, S:semiring=Log):
        raise NotImplementedError

    def viterbi(self, scores):
        raise NotImplementedError

    def ctc_loss(self, scores, targets, target_lengths):
        raise NotImplementedError

    def posteriors(self, scores, S:semiring=Log):
        f = lambda x: self.logZ(x, S).sum()
        return grad(f, scores)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/04_CTC_loss_simple.ipynb (unless otherwise specified).

__all__ = ['device', 'generate_sample_inputs', 'logZ_fwd', 'dot', 'LogZ', 'logZ_py', 'mean', 'logZ_cupy',
           'viterbi_alignments', 'soft_alignments', 'cupy_funcs', 'cupy_funcs_loop']

# Cell
import numpy as np
import cupy as cp
import torch
import torch.nn as nn
from collections import namedtuple
from .utils import *

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

# Cell
def generate_sample_inputs(T, N, L_min, L_max, device=device):
    """
    Args:
        T: number of time steps
        N: batch size
        L_min, L_max: bounds on target length
    """
    stay_scores = torch.rand(T, N, L_max, device=device, requires_grad=True)
    move_scores = torch.rand(T, N, L_max-1, device=device, requires_grad=True)
    target_lengths = torch.randint(L_min, L_max+1, (N,), device=device)
    return stay_scores, move_scores, target_lengths

# Cell
from torch.nn.functional import pad
from .core import semiring, Log, Max

def logZ_fwd(stay_scores, move_scores, target_lengths, S=Log):
    T, N, L = stay_scores.shape
    alpha_0 = stay_scores.new_full((N, L), S.zero); alpha_0[:, 0] = S.one
    beta_T = stay_scores.new_full((N, L), S.zero); beta_T[torch.arange(N), target_lengths - 1] = S.one
    move_scores = pad(move_scores, (1, 0), value=S.zero)
    a = pad(alpha_0, (1, 0), value=S.zero)
    for t in range(0, stay_scores.size(0)):
        a[:, 1:] = S.sum(torch.stack([
            S.mul(stay_scores[t], a[:, 1:]),
            S.mul(move_scores[t], a[:, :-1])
        ]), dim=0)
    return S.sum(S.mul(a[:, 1:], beta_T), dim=1)

# Cell
def _simple_lattice_fwd_bwd(alpha, beta_T, beta_stay, beta_move, stay_scores, move_scores, S=Log):
    T = alpha.size(0) - 1
    move_scores = pad(move_scores, (1, 1), value=S.zero)
    a = pad(alpha[0], (1, 0), value=S.zero)
    for t in range(0, T):
        a[:, 1:] = S.sum(torch.stack([
            S.mul(stay_scores[t], a[:, 1:]),
            S.mul(move_scores[t, :, :-1], a[:, :-1])
        ]), dim=0)
        alpha[t+1] = a[:, 1:]

    b = pad(beta_T, (0, 1), value=S.zero)
    for t in range(T, 0, -1):
        beta_stay[t-1] = S.mul(b[:, :-1], stay_scores[t - 1])
        beta_move[t-1] = S.mul(b[:, 1:], move_scores[t - 1, :, 1:])
        b[:, :-1] = S.sum(torch.stack([beta_stay[t-1], beta_move[t-1]]), dim=0)

def dot(x, y, S=Log, dim=-1):
    return S.sum(S.mul(x, y), dim=dim)

class LogZ(torch.autograd.Function):
    @staticmethod
    def forward(ctx, stay_scores, move_scores, target_lengths, fwd_bwd_impl, S:semiring):
        T, N, L = stay_scores.shape

        alpha = stay_scores.new_full((T + 1, N, L), S.zero)
        alpha[0, :, 0] = S.one

        beta_stay = stay_scores.new_full((T, N, L), S.zero)
        beta_move = stay_scores.new_full((T, N, L), S.zero)
        beta_T = stay_scores.new_full((N, L), S.zero)
        beta_T[torch.arange(N), target_lengths - 1] = S.one

        fwd_bwd_impl(alpha, beta_T, beta_stay, beta_move, stay_scores, move_scores, S)

        g = S.dsum(torch.cat([S.mul(alpha[:-1], beta_stay), S.mul(alpha[:-1], beta_move)], dim=2), dim=2)

        ctx.save_for_backward(g.reshape(T, N, 2, L))
        return dot(alpha[-1], beta_T, S)

    @staticmethod
    def backward(ctx, grad):
        g = ctx.saved_tensors[0] * grad[None, :, None, None]
        return g[:, :, 0], g[:, :, 1, :-1], None, None, None

def logZ_py(stay_scores, move_scores, target_lengths):
    return LogZ.apply(stay_scores, move_scores, target_lengths, _simple_lattice_fwd_bwd, Log)

# Cell
mean = lambda f: (lambda *xs: f(*xs).mean())

# Cell
from .utils import *
import cupy as cp

cupy_funcs = {
    (torch.float32, Log): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace', FLOAT='float',  SUM='logsumexp2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float64, Log): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace', FLOAT='double', SUM='logsumexp2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float32, Max): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace', FLOAT='float',  SUM='max2', MUL='add', ZERO='{:E}'.format(Max.zero)),
    (torch.float64, Max): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace', FLOAT='double', SUM='max2', MUL='add', ZERO='{:E}'.format(Max.zero)),
}

cupy_funcs_loop = {
    (torch.float32, Log): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace_loop', FLOAT='float',  SUM='logsumexp2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float64, Log): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace_loop', FLOAT='double', SUM='logsumexp2', MUL='add', ZERO='{:E}'.format(Log.zero)),
    (torch.float32, Max): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace_loop', FLOAT='float',  SUM='max2', MUL='add', ZERO='{:E}'.format(Max.zero)),
    (torch.float64, Max): load_cupy_func('cuda/ctc_simple.cu', 'fwd_bwd_logspace_loop', FLOAT='double', SUM='max2', MUL='add', ZERO='{:E}'.format(Max.zero)),
}

def _simple_lattice_fwd_bwd_cupy(alpha, beta_T, beta_stay, beta_move, stay_scores, move_scores, S:semiring):
    T, N, L = stay_scores.shape
    if L > 1024: #exceeds max threads per block
        return _simple_lattice_fwd_bwd_cupy_loop(alpha, beta_T, beta_stay, beta_move, stay_scores, move_scores, S)
    _bytes = 8 if (stay_scores.dtype == torch.float64) else 4
    with cp.cuda.Device(stay_scores.device.index):
        cupy_funcs[(stay_scores.dtype, S)](grid=(N, 2, 1), block=(L, 1, 1), shared_mem=2*_bytes*L,
               args=(alpha.data_ptr(), beta_T.data_ptr(), beta_stay.data_ptr(), beta_move.data_ptr(),
                     stay_scores.data_ptr(), move_scores.data_ptr(), T, N, L))

def _simple_lattice_fwd_bwd_cupy_loop(alpha, beta_T, beta_stay, beta_move, stay_scores, move_scores, S:semiring, max_block_size=1024):
    T, N, L = stay_scores.shape
    block_size = min(L, max_block_size)
    beta = alpha.new_full(alpha.shape, S.zero)
    beta[-1] = beta_T
    with cp.cuda.Device(stay_scores.device.index):
        cupy_funcs_loop[(stay_scores.dtype, S)](grid=(N, 2, 1), block=(block_size, 1, 1),
               args=(alpha.data_ptr(), beta.data_ptr(), beta_stay.data_ptr(), beta_move.data_ptr(),
                     stay_scores.data_ptr(), move_scores.data_ptr(), T, N, L))

def logZ_cupy(stay_scores, move_scores, target_lengths, S:semiring=Log):
    return LogZ.apply(stay_scores, move_scores, target_lengths, _simple_lattice_fwd_bwd_cupy, S)

def viterbi_alignments(stay_scores, move_scores, target_lengths):
    target_lengths = target_lengths.to(stay_scores.device)
    stay_scores, move_scores = stay_scores.detach().requires_grad_(), move_scores.detach().requires_grad_()
    logZ_cupy(stay_scores, move_scores, target_lengths, Max).sum().backward()
    alignments = stay_scores.grad.clone()
    alignments[:, :, :-1] += move_scores.grad
    return alignments

def soft_alignments(stay_scores, move_scores, target_lengths, beta=1.0):
    target_lengths = target_lengths.to(stay_scores.device)
    if beta != 1.0:
        stay_scores, move_scores = stay_scores * beta, move_scores * beta
    stay_scores, move_scores = stay_scores.detach().requires_grad_(), move_scores.detach().requires_grad_()
    logZ_cupy(stay_scores, move_scores, target_lengths, Log).sum().backward()
    alignments = stay_scores.grad.clone()
    alignments[:, :, :-1] += move_scores.grad
    return alignments


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/05_Normalisation.ipynb (unless otherwise specified).

__all__ = ['device', 'generate_test_example', 'logZ_py', 'fused_batch_Mv', 'LogZ', 'LogZViterbi', 'logz', 'cupy_funcs']

# Cell
import numpy as np
import cupy as cp
import torch
from .utils import *

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

# Cell
def generate_test_example(T, N, n_state, dtype=torch.float):
    return torch.rand((T, N, n_state, n_state), device=device, dtype=dtype, requires_grad=True)

# Cell
import torch

def _rescale(M):
    #T, N, n_state, n_state = M.shape
    Z = M.sum((2, 3), keepdim=True) / M.size(3)
    logZ = torch.log(Z).sum(0).reshape(-1)
    return M / Z, logZ

@torch.jit.script
def logZ_py(M, alpha_0):
    M, logZ = _rescale(M)
    T, N, n_state, _ = M.shape
    alpha = alpha_0.unsqueeze(2)
    for i, M_t in enumerate(M.unbind(0)):
        alpha = M_t.bmm(alpha)
        if i % 32 == (T - 1) % 32:
            z = alpha.sum(1, keepdim=True)
            alpha = alpha/z
            logZ += torch.log(z.squeeze())
    return logZ

# Cell
from .core import semiring, Log, Max
from functools import partial

cupy_funcs = {
    (torch.float32, Log): load_cupy_func('cuda/fused_bmv.cu', 'fwd', FLOAT='float', MUL='add', SUM='logsumexp2'),
    (torch.float64, Log): load_cupy_func('cuda/fused_bmv.cu', 'fwd', FLOAT='double', MUL='add', SUM='logsumexp2'),
    (torch.float32, Max): load_cupy_func('cuda/fused_bmv.cu', 'fwd', FLOAT='float',  MUL='add', SUM='max2'),
    (torch.float64, Max): load_cupy_func('cuda/fused_bmv.cu', 'fwd', FLOAT='double', MUL='add', SUM='max2'),
}

def fused_batch_Mv(Ms, alpha_0, S:semiring=Log):
    T, N, n_state, _ = Ms.shape
    alpha = Ms.new_empty((T + 1, N, n_state))
    alpha[0] = alpha_0
    with cp.cuda.Device(Ms.device.index):
        cupy_funcs[(Ms.dtype, S)](
            grid=(N, 1, 1),
            block=(n_state, 1, 1),
            args=(alpha.data_ptr(), Ms.contiguous().data_ptr(), T, N, n_state)
        )
    return alpha

def _logz_fwd(ctx, Ms, alpha_0, beta_T, S:semiring=Log):
    alpha = fused_batch_Mv(Ms, alpha_0, S)
    ctx.save_for_backward(Ms, alpha, beta_T)
    return S.sum(S.mul(alpha[-1], beta_T), dim=1)

def _logz_bwd(ctx, g, S:semiring=Log):
    Ms, alpha, beta_T = ctx.saved_tensors
    T, N, n_state, _ = Ms.shape
    beta = fused_batch_Mv(Ms.transpose(2, 3).flip(0), beta_T, S)
    Ms_grad = S.mul(S.mul(Ms, alpha[:-1,:,None,:]), (beta[:-1, :, :, None]).flip(0))
    Ms_grad = S.dsum(Ms_grad.reshape(T, N, -1), dim=2).reshape(T, N, n_state, n_state)
    return Ms_grad * g[None, :, None, None], None, None, None

class LogZ(torch.autograd.Function):
    forward = staticmethod(_logz_fwd)
    backward = staticmethod(_logz_bwd)

class LogZViterbi(torch.autograd.Function):
    forward = staticmethod(partial(_logz_fwd, S=Max))
    backward = staticmethod(partial(_logz_bwd, S=Max))

def logz(Ms, alpha_0, beta_T, S:semiring=Log):
    if S==Log:
        return LogZ.apply(Ms, alpha_0, beta_T)
    elif S==Max:
        return LogZViterbi.apply(Ms, alpha_0, beta_T)
    else:
        raise Exception('semiring {} not supported'.format(S))